/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.binding.http;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.xml.XMLConstants;
import javax.xml.namespace.QName;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;
import javax.xml.validation.Validator;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.Resource;
import org.bluestemsoftware.specification.eoa.component.ComponentContext;
import org.bluestemsoftware.specification.eoa.component.ComponentDeployment;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentType;
import org.bluestemsoftware.specification.eoa.component.binding.Binding;
import org.bluestemsoftware.specification.eoa.component.binding.rt.BindingRT;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction.Direction;
import org.bluestemsoftware.specification.eoa.ext.binding.BindingFactory;
import org.bluestemsoftware.specification.eoa.ext.binding.http.rt.HTTPBindingConfiguration;
import org.bluestemsoftware.specification.eoa.ext.binding.http.rt.HTTPBindingRT;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HTTPHeader;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HyperTextTransferProtocol;
import org.bluestemsoftware.specification.eoa.ext.idl.IDLDocument;
import org.bluestemsoftware.specification.eoa.ext.idl.IDLException;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.WSDLException;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl20.WSDL20Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
 * A type specific <code>Binding</code> factory that creates HTTP bindings.
 */
public class HTTPBindingFactory extends BindingFactory {

    /*
     * (non-Javadoc)
     * 
     * SPI
     */
    public interface Provider extends BindingFactory.Provider {

        public HTTPBindingRT.Provider spi_createDefaultProvider(HTTPBindingFactory factory, ComponentName componentName, String protocolType) throws DeploymentException;

        public HTTPBindingRT.Provider spi_createProvider(HTTPBindingFactory factory, ComponentName componentName, String protocolType, HTTPBindingConfiguration hbc) throws DeploymentException;

    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.ExtensionFactory#getExtensionType()
     */
    public String getExtensionType() {
        return HTTPBinding.TYPE;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.binding.BindingFactory#getBindingType()
     */
    @Override
    public String getBindingType() {
        return HTTPBinding.BINDING_TYPE;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.binding.BindingFactory#createDefaultProvider(org.bluestemsoftware.specification.eoa.component.ComponentContext,
     *      org.bluestemsoftware.specification.eoa.component.binding.Binding)
     */
    @Override
    public HTTPBindingRT createDefaultBinding(ComponentContext componentContext, Binding runtimeProvidable) throws DeploymentException {

        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();

        try {

            thread.setContextClassLoader(factoryContext.getClassLoader());

            if (!(runtimeProvidable instanceof HTTPBinding)) {
                throw new DeploymentException("Expected instanceof " + HTTPBinding.class.getName());
            }

            ComponentName componentName = runtimeProvidable.getName();
            String pt = runtimeProvidable.getUnderlyingProtocol().getProtocolType();
            HTTPBindingRT.Provider p = ((Provider)provider).spi_createDefaultProvider(this, componentName, pt);

            if (pt.equals(XMLHTTPProtocol.TYPE)) {
                return new HTTPBindingRT(this, (HTTPBindingRT.Provider)p, componentName, pt);
            } else {
                throw new DeploymentException("Error creating default provider definition "
                        + componentName
                        + ". Corresponding Binding employs an unsupported protocol "
                        + pt
                        + ".");
            }

        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.idl.wsdl.WSDLFactory$BindingDefinitionReader#readBinding(org.bluestemsoftware.specification.eoa.component.ComponentContext,
     *      java.lang.String, org.w3c.dom.Element,
     *      org.bluestemsoftware.specification.eoa.ext.idl.wsdl.WSDLFactory.BindingDefinitionReader.BindingStyle)
     */
    public void readBinding(ComponentContext componentContext, String targetNamespace, Element bindingElement, BindingStyle bindingStyle) throws DeploymentException {

        if (factoryContext == null) {
            throw new IllegalStateException("Factory has no context.");
        }
        if (componentContext == null) {
            throw new IllegalArgumentException("componentContext null");
        }
        if (targetNamespace == null) {
            throw new IllegalArgumentException("targetNamespace null");
        }
        if (bindingElement == null) {
            throw new IllegalArgumentException("bindingElement null");
        }

        // note that we only allow definition of http binding via wsdl 2.0. b/c this is
        // the spec that defines http binding extension. basic profile prohibits binding
        // defs within wsdl 1.1 other than soap bindings

        Binding binding = null;
        if (bindingElement.getNamespaceURI().equals(WSDL20Document.TYPE)) {
            binding = org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl20.ext.HTTPBindingReader
                    .readBinding(
                            this,
                            componentContext,
                            targetNamespace,
                            (org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl20.BindingElement)bindingElement,
                            bindingStyle);
        } else {
            throw new DeploymentException("Unsupported wsdl version " + bindingElement.getNamespaceURI() + ".");
        }

        componentContext.getDeployment().addComponent(binding);

    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.idl.wsdl.WSDLFactory$BindingWriter#writeBinding(org.bluestemsoftware.specification.eoa.ext.idl.wsdl.WSDLDocument,
     *      org.bluestemsoftware.specification.eoa.component.binding.Binding)
     */
    public void writeBinding(IDLDocument idlDocument, Binding binding) throws IDLException {
        if (idlDocument == null) {
            throw new IllegalArgumentException("idlDocument null");
        }
        if (binding == null) {
            throw new IllegalArgumentException("binding null");
        }

        // note that we only export definition of http binding via wsdl 2.0. b/c this is
        // the spec that defines http binding extension. basic profile prohibits binding
        // defs within wsdl 1.1 other than soap bindings

        if (idlDocument instanceof WSDL20Document) {
            org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl20.ext.HTTPBindingWriter.writeBinding(
                    (WSDL20Document)idlDocument, (HTTPBinding)binding);
        } else {
            throw new WSDLException("Unsupported idlDocument type " + idlDocument.getClass().getName() + ".");
        }

    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.ProviderReader#readProviders(org.bluestemsoftware.specification.eoa.component.ComponentContext,
     *      javax.xml.transform.dom.DOMSource)
     */
    public Set<BindingRT> readProviders(ComponentContext componentContext, DOMSource source) throws DeploymentException {

        if (componentContext == null) {
            throw new IllegalArgumentException("entityResolver null");
        }
        if (source == null) {
            throw new IllegalArgumentException("source null");
        }

        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {

            thread.setContextClassLoader(factoryContext.getClassLoader());

            validateDefinition(source);

            ComponentDeployment deployment = componentContext.getDeployment();

            Element definition = (Element)source.getNode();
            String prefixedName = definition.getAttribute("name");
            ComponentName cname = null;
            if (prefixedName.equals("")) {
                throw new DeploymentException("Error loading runtime provider definition "
                        + ". Definition missing required attribute 'name'.");
            } else {
                cname = ComponentName.valueOf(definition, definition.getAttribute("name"));
            }

            Binding binding = (Binding)deployment.getComponent(ComponentType.BINDING, cname);

            if (binding == null) {
                throw new DeploymentException("Error loading runtime provider definition "
                        + cname
                        + ". Definition for corresponding RootComponent is undefined.");
            }

            if (!(binding instanceof HTTPBinding)) {
                throw new DeploymentException("Error loading runtime provider definition "
                        + cname
                        + ". Corresponding RootComponent is not an instance of "
                        + HTTPBinding.class);
            }

            QName childName = new QName(HTTPBinding.TYPE, "configuration");
            Element configurationElement = getChildElement(definition, childName);

            HTTPBindingConfiguration hbc = new HTTPBindingConfiguration();

            if (configurationElement != null) {

                XMLHTTPProtocol hup = (XMLHTTPProtocol)binding.getUnderlyingProtocol();

                childName = new QName(HTTPBinding.TYPE, "faults");
                Element faultsElement = getChildElement(configurationElement, childName);
                if (faultsElement != null) {
                    childName = new QName(HTTPBinding.TYPE, "fault");
                    List<Element> list = getChildElements(faultsElement, childName);
                    for (Element faultElement : list) {
                        prefixedName = faultElement.getAttribute("name");
                        QName faultName = ComponentName.valueOf(faultElement, prefixedName);
                        childName = new QName(HTTPBinding.TYPE, "httpHeaders");
                        Map<QName, Map<String, String>> faultHeaderValues = null;
                        Element headersElement = getChildElement(faultElement, childName);
                        if (headersElement == null) {
                            continue;
                        } else {
                            faultHeaderValues = new HashMap<QName, Map<String, String>>();
                            hbc.setFaultHeaderValues(faultHeaderValues);
                        }
                        childName = new QName(HTTPBinding.TYPE, "httpHeader");
                        List<Element> headerElements = getChildElements(headersElement, childName);
                        for (Element headerElement : headerElements) {
                            Map<String, String> headerMap = faultHeaderValues.get(faultName);
                            if (headerMap == null) {
                                headerMap = new HashMap<String, String>();
                                faultHeaderValues.put(faultName, headerMap);
                            }
                            String headerName = headerElement.getAttribute("name");
                            Collection<HTTPHeader> decls = ((HyperTextTransferProtocol)hup)
                                    .getFaultHTTPHeaders(faultName);
                            boolean found = false;
                            for (HTTPHeader headerDecl : decls) {
                                if (headerDecl.getName().equals(headerName)) {
                                    found = true;
                                    break;
                                }
                            }
                            if (!found) {
                                throw new DeploymentException("Error loading runtime provider definition "
                                        + cname
                                        + ". Value defined for http header '"
                                        + headerName
                                        + "' defined for binding fault '"
                                        + faultName
                                        + "' is not declared on corresponding binding.");
                            }
                            if (headerMap.get(headerName) != null) {
                                throw new DeploymentException("Error loading runtime provider definition "
                                        + cname
                                        + ". Duplicate value defined for http header '"
                                        + headerName
                                        + "' defined for binding fault '"
                                        + faultName
                                        + "'.");
                            } else {
                                headerMap.put(headerName, getText(headerElement));
                            }
                        }
                    }
                }

                childName = new QName(HTTPBinding.TYPE, "operations");
                Element operationsElement = getChildElement(configurationElement, childName);
                if (operationsElement != null) {

                    childName = new QName(HTTPBinding.TYPE, "operation");
                    Map<QName, Map<Direction, Map<String, String>>> messageHeaderValues = null;
                    Map<QName, Integer> requestorTimeout = new HashMap<QName, Integer>();
                    List<Element> list = getChildElements(operationsElement, childName);
                    for (Element operationElement : list) {

                        prefixedName = operationElement.getAttribute("name");
                        QName operationName = ComponentName.valueOf(operationElement, prefixedName);
                        if (binding.getOperation(operationName) == null) {
                            throw new DeploymentException("Error loading runtime provider definition "
                                    + cname
                                    + ". Operation element defined within configuration for operation "
                                    + operationName
                                    + ", but corresponding binding defines no such operation.");
                        }

                        childName = new QName(HTTPBinding.TYPE, "input");
                        Element inputElement = getChildElement(operationElement, childName);
                        if (inputElement != null) {
                            childName = new QName(HTTPBinding.TYPE, "httpHeaders");
                            Element headersElement = getChildElement(inputElement, childName);
                            if (headersElement == null) {
                                continue;
                            }
                            if (messageHeaderValues == null) {
                                messageHeaderValues = new HashMap<QName, Map<Direction, Map<String, String>>>();
                                hbc.setMessageHeaderValues(messageHeaderValues);
                            }
                            Map<Direction, Map<String, String>> operationHeaders = messageHeaderValues
                                    .get(operationName);
                            if (operationHeaders == null) {
                                operationHeaders = new HashMap<Direction, Map<String, String>>();
                                messageHeaderValues.put(operationName, operationHeaders);
                            }
                            childName = new QName(HTTPBinding.TYPE, "httpHeader");
                            List<Element> headerElements = getChildElements(headersElement, childName);
                            for (Element headerElement : headerElements) {
                                Map<String, String> headerMap = operationHeaders.get(Direction.IN);
                                if (headerMap == null) {
                                    headerMap = new HashMap<String, String>();
                                    operationHeaders.put(Direction.IN, headerMap);
                                }
                                String headerName = headerElement.getAttribute("name");
                                Collection<HTTPHeader> decls = ((HyperTextTransferProtocol)hup)
                                        .getMessageReferenceHTTPHeaders(operationName, Direction.IN);
                                boolean found = false;
                                for (HTTPHeader headerDecl : decls) {
                                    if (headerDecl.getName().equals(headerName)) {
                                        found = true;
                                        break;
                                    }
                                }
                                if (!found) {
                                    throw new DeploymentException("Error loading runtime provider definition "
                                            + cname
                                            + ". Value defined for http header '"
                                            + headerName
                                            + "' defined for input message on operation '"
                                            + operationName
                                            + "' is not declared on corresponding binding.");
                                }
                                if (headerMap.get(headerName) != null) {
                                    throw new DeploymentException("Error loading runtime provider definition "
                                            + cname
                                            + ". Duplicate value defined for http header '"
                                            + headerName
                                            + "' defined on input message for binding operation '"
                                            + operationName
                                            + "'.");
                                } else {
                                    headerMap.put(headerName, getText(headerElement));
                                }
                            }
                        }

                        childName = new QName(HTTPBinding.TYPE, "output");
                        Element outputElement = getChildElement(operationElement, childName);
                        if (outputElement != null) {
                            childName = new QName(HTTPBinding.TYPE, "httpHeaders");
                            Element headersElement = getChildElement(outputElement, childName);
                            if (headersElement == null) {
                                continue;
                            }
                            if (messageHeaderValues == null) {
                                messageHeaderValues = new HashMap<QName, Map<Direction, Map<String, String>>>();
                                hbc.setMessageHeaderValues(messageHeaderValues);
                            }
                            Map<Direction, Map<String, String>> operationHeaders = messageHeaderValues
                                    .get(operationName);
                            if (operationHeaders == null) {
                                operationHeaders = new HashMap<Direction, Map<String, String>>();
                                messageHeaderValues.put(operationName, operationHeaders);
                            }
                            childName = new QName(HTTPBinding.TYPE, "httpHeader");
                            List<Element> headerElements = getChildElements(headersElement, childName);
                            for (Element headerElement : headerElements) {
                                Map<String, String> headerMap = operationHeaders.get(Direction.OUT);
                                if (headerMap == null) {
                                    headerMap = new HashMap<String, String>();
                                    operationHeaders.put(Direction.OUT, headerMap);
                                }
                                String headerName = headerElement.getAttribute("name");
                                Collection<HTTPHeader> decls = ((HyperTextTransferProtocol)hup)
                                        .getMessageReferenceHTTPHeaders(operationName, Direction.OUT);
                                boolean found = false;
                                for (HTTPHeader headerDecl : decls) {
                                    if (headerDecl.getName().equals(headerName)) {
                                        found = true;
                                        break;
                                    }
                                }
                                if (!found) {
                                    throw new DeploymentException("Error loading runtime provider definition "
                                            + cname
                                            + ". Value defined for http header '"
                                            + headerName
                                            + "' defined for output message on operation '"
                                            + operationName
                                            + "' is not declared on corresponding binding.");
                                }
                                if (headerMap.get(headerName) != null) {
                                    throw new DeploymentException("Error loading runtime provider definition "
                                            + cname
                                            + ". Duplicate value defined for http header '"
                                            + headerName
                                            + "' defined on output message for binding operation '"
                                            + operationName
                                            + "'.");
                                } else {
                                    headerMap.put(headerName, getText(headerElement));
                                }
                            }
                        }

                        childName = new QName(HTTPBinding.TYPE, "requestorTimeout");
                        Element timeoutElement = getChildElement(operationElement, childName);
                        if (timeoutElement != null) {
                            Integer timeout = Integer.valueOf(getText(timeoutElement));
                            requestorTimeout.put(operationName, timeout);
                        }

                    }
                    hbc.setRequestorTimeout(requestorTimeout);
                }

                childName = new QName(HTTPBinding.TYPE, "dataSourceConnector");
                Element dsElement = getChildElement(configurationElement, childName);
                if (dsElement != null) {
                    hbc.setDataSourceConnectorName(getText(dsElement));
                }

                childName = new QName(HTTPBinding.TYPE, "embedStackTraceInFaults");
                Element embedElement = getChildElement(configurationElement, childName);
                if (embedElement != null) {
                    hbc.setEmbeddStackTraceInFaults(Boolean.parseBoolean(getText(embedElement)));
                }

                childName = new QName(HTTPBinding.TYPE, "maxPayloadProcessors");
                Element payloadElement = getChildElement(configurationElement, childName);
                if (payloadElement != null) {
                    int value = Integer.parseInt(getText(payloadElement));
                    hbc.setMaxPayloadProcessors(value);
                }
            }

            String pt = binding.getUnderlyingProtocol().getProtocolType();
            Set<BindingRT> providers = new HashSet<BindingRT>();

            HTTPBindingRT bindingRT = null;
            if (pt.equals(XMLHTTPProtocol.TYPE)) {
                HTTPBindingRT.Provider p = ((Provider)provider).spi_createProvider(this, cname, pt, hbc);
                bindingRT = new HTTPBindingRT(this, (HTTPBindingRT.Provider)p, cname, pt);
            } else {
                throw new DeploymentException("Error creating provider definition "
                        + cname
                        + ". Corresponding Binding employs an unsupported protocol "
                        + pt
                        + ".");
            }

            providers.add(bindingRT);
            return providers;

        } finally {
            thread.setContextClassLoader(cl);
        }

    }

    private Element getChildElement(Element parent, QName childName) {
        NodeList nodeList = parent.getChildNodes();
        for (int i = 0; i < nodeList.getLength(); i++) {
            if (nodeList.item(i).getNodeType() != Node.ELEMENT_NODE) {
                continue;
            }
            Element element = (Element)nodeList.item(i);
            if (childName.getNamespaceURI().equals(XMLConstants.NULL_NS_URI)) {
                if (element.getNamespaceURI() != null) {
                    continue;
                }
            } else {
                if (element.getNamespaceURI() == null) {
                    continue;
                }
                if (!childName.getNamespaceURI().equals("*")
                        && !element.getNamespaceURI().equals(childName.getNamespaceURI())) {
                    continue;
                }
            }
            if (!childName.getLocalPart().equals("*") && !element.getLocalName().equals(childName.getLocalPart())) {
                continue;
            }
            return element;
        }
        return null;
    }

    private List<Element> getChildElements(Element parent, QName childName) {
        List<Element> answer = new ArrayList<Element>();
        NodeList nodeList = parent.getChildNodes();
        for (int i = 0; i < nodeList.getLength(); i++) {
            if (nodeList.item(i).getNodeType() != Node.ELEMENT_NODE) {
                continue;
            }
            Element element = (Element)nodeList.item(i);
            if (childName.getNamespaceURI().equals(XMLConstants.NULL_NS_URI)) {
                if (element.getNamespaceURI() != null) {
                    continue;
                }
            } else {
                if (element.getNamespaceURI() == null) {
                    continue;
                }
                if (!childName.getNamespaceURI().equals("*")
                        && !element.getNamespaceURI().equals(childName.getNamespaceURI())) {
                    continue;
                }
            }
            if (!childName.getLocalPart().equals("*") && !element.getLocalName().equals(childName.getLocalPart())) {
                continue;
            }
            answer.add(element);
        }
        return answer;
    }

    private void validateDefinition(DOMSource source) throws DeploymentException {

        Resource resource = null;
        try {
            resource = factoryContext.getResource(HTTPBinding.PROVIDER_SCHEMA_LOCATION);
        } catch (DeploymentException de) {
            throw new DeploymentException("Error validating definition. " + de);
        }

        SchemaFactory schemaFactory = SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);

        javax.xml.validation.Schema schema = null;
        try {
            schema = schemaFactory.newSchema(new StreamSource(resource.getInputStream()));
        } catch (Exception ex) {
            throw new DeploymentException("Error parsing schema. " + ex);
        }
        DOMValidator domValidator = new DOMValidator(schema);
        String validationError;
        try {
            validationError = domValidator.validate(source);
        } catch (Exception ex) {
            throw new DeploymentException("Error validating definition. " + ex);
        }
        if (validationError != null) {
            throw new DeploymentException("Engine definition invalid. " + validationError);
        }

    }

    private String getText(Element element) {
        if (element == null || !element.hasChildNodes()) {
            return null;
        }
        try {
            Text textNode = (Text)element.getFirstChild();
            return textNode.getData();
        } catch (ClassCastException ce) {
        }
        return null;
    }

    static final class DOMValidator {

        private Schema schema;

        public DOMValidator(Schema schema) {
            this.schema = schema;
        }

        public String validate(DOMSource source) throws SAXException, IOException {
            Validator validator = schema.newValidator();
            validator.setErrorHandler(new ErrorHandlerImpl());
            validator.validate(source);
            return ((ErrorHandlerImpl)validator.getErrorHandler()).getErrors();
        }

        static class ErrorHandlerImpl implements ErrorHandler {

            private ArrayList<String> errorMessages = new ArrayList<String>();

            public void error(SAXParseException spe) throws SAXException {
                errorMessages.add(spe.getMessage());
            }

            public void warning(SAXParseException spe) throws SAXException {
                errorMessages.add(spe.getMessage());
            }

            public void fatalError(SAXParseException spe) throws SAXException {
                errorMessages.add(spe.getMessage());
            }

            public String getErrors() {
                if (errorMessages.size() > 0) {
                    Iterator<String> messageIterator = errorMessages.iterator();
                    StringBuilder errorMessage = new StringBuilder();
                    while (messageIterator.hasNext()) {
                        errorMessage.append(messageIterator.next() + "  ");
                    }
                    return errorMessage.toString();
                }
                return null;
            }

        }

    }

}
