/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.binding.http.rt;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.binding.rt.BindingRT;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointReference;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction.Direction;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.MessageContext;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.SystemFault;
import org.bluestemsoftware.specification.eoa.component.message.rt.Message;
import org.bluestemsoftware.specification.eoa.ext.binding.BindingFactory;
import org.bluestemsoftware.specification.eoa.ext.feature.http.server.HTTPServerRequest;
import org.bluestemsoftware.specification.eoa.ext.feature.http.server.HTTPServerResponse;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.rt.HyperTextTransferProtocolRT;

/**
 * A runtime binding which processes 'plain old xml' messages over HTTP at runtime.
 */
public final class HTTPBindingRT extends BindingRT implements HyperTextTransferProtocolRT {

    /*
     * (non-Javadoc)
     * 
     * SPI
     */
    public interface Provider extends BindingRT.Provider {

        public void spi_sendAction(EndpointReference endpointReference, ActionContext actionContext) throws SystemFault;

        public void spi_receiveAction(HTTPServerRequest request, HTTPServerResponse response, Direction direction, EndpointReference endpointReference);

    }

    public HTTPBindingRT(BindingFactory factory, Provider provider, ComponentName componentName,
            String protocolType) {
        super(factory, provider, componentName, protocolType);
    }

    /**
     * Invoked by an instance of <code>HTTPTransportModule</code>. Receives action addressed
     * to referenced endpoint, i.e. referenced Service upon which action is defined is 'my'
     * Service, action represents a request and message direction is IN.
     * <p>
     * Note that (despite the fact that this binding does not use the SOAP protocol) this
     * extension uses the SOAP 1.2 'state machine' terminology along with requirements defined
     * within version 1.2 of the basic profile and the WSDL 2.0 adjuncts specification to
     * define the expected behavior of a compliant HTTP binding when receiving an action.
     * <p>
     * The action MUST be considered as 'accepted for processing' by this binding, i.e. the
     * 'responding soap node', once the binding transitions from the 'init' state to the
     * 'receiving' state. If the action is not 'accepted for processing', a 4xx HTTP status
     * code and an empty response body MUST be returned.
     * <p>
     * If the action is 'accepted for processing' and WSDL MEP is 'in-only', a 202 HTTP status
     * code along with an empty response body MUST be returned. If this binding, i.e. the
     * 'responding node' fails during the 'receiving' state for any reason while processing a
     * WSDL 'in-only' MEP, the error MUST NOT be communicated to the 'requesting node'. It
     * SHOULD, however, be logged.
     * <p>
     * If the action is 'accepted for processing' and WSDL MEP is 'robust-in-only ' and the
     * request was successful, a 204 HTTP status code along with an empty response body MUST be
     * returned.
     * <p>
     * If the action is 'accepted for processing' and WSDL MEP is 'in-out' or 'robust-in-only '
     * and this binding, i.e. the 'responding node' transitions from the 'receiving' state to
     * the 'receiving+sending' or 'sending' state, the response MUST be returned within the
     * HTTP response body. If the response is a fault, the status code MUST be 500 (when
     * binding is serialized via WSDL 2.0, each binding fault declaration SHOULD define this as
     * the value of the whttp:code attribute.) If response is a normal response, i.e. the WSDL
     * MEP is 'in-out', the status code returned MUST be 200.
     * 
     * @param request
     * @param response
     * @param endpointReference
     * 
     * @see http://www.w3.org/TR/wsdl20-adjuncts/#http-binding
     * @see http://www.w3.org/TR/2003/REC-soap12-part2-20030624/#http-msgexop
     * @see http://www.ws-i.org/Profiles/BasicProfile-1_2(WGAD).html
     */
    public void receiveAction(HTTPServerRequest request, HTTPServerResponse response, Direction direction, EndpointReference endpointReference) {
        if (provider == null) {
            throw new IllegalStateException("provider null");
        }
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            ((Provider)provider).spi_receiveAction(request, response, direction, endpointReference);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /**
     * Sends action to referenced <code>Endpoint</code>, i.e. referenced Service upon which
     * action is defined is a 'partner' Service, action represents a request and message
     * direction is IN.
     * <p>
     * Note that if WSDL MEP is either 'in-out' or 'robust-in-only' and a message is present
     * within the HTTP response body, the response MUST be returned by Binding to engine via
     * the method
     * {@link #receiveAction(HttpServletRequest, HttpServletResponse, EndpointReference)}.
     * <p>
     * Note that if WSDL MEP is 'in-only' and a message is present within the HTTP response
     * body, the response MUST be discarded by a compliant Binding provider.
     * 
     * @param ActionContext
     *        Defines context within which action occurs. Message direction MUST be IN and
     *        action MUST be an action declared on Interface implemented by 'partner' Service,
     *        i.e. manifested as an instance of {@link MessageContext}.
     * 
     * @throws SystemFault
     *         Thrown by <code>Binding</code> as an indication that action was NOT 'accepted
     *         for processing' by the 'receiver.' Note that this extension uses the SOAP 1.2
     *         'state machine' (despite the fact that this binding does not use the SOAP
     *         protocol) terminology as a reference in determining whether or not an action has
     *         been 'accepted for processing.'
     *         <p>
     *         An action MUST be considered by Binding provider as 'accepted for processing'
     *         once this binding, i.e. the 'requesting node', transitions from the 'requesting'
     *         state to the 'sending+receiving' or 'receiving' state (see the rules defined
     *         above regarding handling of the response, if any.)
     *         <p>
     *         If this Binding, i.e. the 'requesting node' either fails to initiate the action
     *         or if a failure is encountered within the 'requesting' state and the binding
     *         transitions into the 'fail' state, a <code>SystemFault</code> MUST be thrown.
     *         This Binding, i.e. the 'requesting node', MUST transition from the 'requesting'
     *         state to the 'fail' state if the 'responding node' returns an HTTP status line
     *         where the status code is 4xx. This binding, i.e. the 'requesting node' MUST NOT
     *         transition from the 'requesting' state to the 'fail' state for any other reason.
     *         <p>
     *         Note that if an unchecked exception is encountered by Binding before action was
     *         'accepted for processing' by 'responding node', the exception MUST be converted
     *         by Binding to a <code>SystemFault</code> and thrown. If unchecked exception
     *         occurs after the action was 'accepted for processing' by 'responding node', the
     *         exception MUST be converted to a <code>SystemFault</code> and communicated to
     *         engine via the method
     *         {@link #receiveAction(HttpServletRequest, HttpServletResponse, EndpointReference)}.
     *         <p>
     *         Note that a compliant <code>Binding</code> provider MUST NOT throw a
     *         <code>SystemFault</code> for any other reason other than those cited above.
     * 
     * @see http://www.w3.org/TR/wsdl20-adjuncts/#http-binding
     * @see #createAction(String, Message)
     * @see http://www.w3.org/TR/2003/REC-soap12-part2-20030624/#http-msgexop
     */
    @Override
    public void sendAction(EndpointReference endpointReference, ActionContext actionContext) throws SystemFault {
        if (provider == null) {
            return;
        }
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            ((Provider)provider).spi_sendAction(endpointReference, actionContext);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

}
