/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.binding.soap;

import java.util.Map;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.component.binding.BindingOperation;
import org.bluestemsoftware.specification.eoa.component.policy.rt.ElementPolicy;

public final class SOAPBindingOperation extends BindingOperation {

    private SOAPMessageExchangePattern soapMessageExchangePattern;
    private String soapAction;

    /**
     * Constructs a protocol specific, i.e. SOAP, <code>BindingOperation</code> component.
     * 
     * @param name
     *        Qualified name of operation which MUST be unique within context of
     *        <code>Binding</code>.
     * @param inputMessageReference
     *        References concrete message received by operation. Corresponding
     *        InterfaceMessageReference direction is "In".
     * @param outputMessageReference
     *        References concrete message sent by operation. Corresponding
     *        InterfaceMessageReference direction is "Out".
     * @param faultReferences
     *        Collection of components each of which references a <code>BindingFault</code>
     *        component or an empty collection if none defined.
     * @param elementPolicy
     *        Contains policy expressions which may be attached to
     *        <code>BindingOperation</code>.
     * @param soapMessageExchangePattern
     *        Identifies protocol specific, i.e. SOAP, message exchange pattern utilized by
     *        this operation. If {@link SOAPVersion} employed by parent <code>Binding</code>
     *        is soap 1.1, no assertion is made regarding soap message exchange pattern, i.e.
     *        argument MUST be null.
     * @param soapAction
     *        Identifies value of SOAP Action for the initial message of the message exchange
     *        pattern of the bound interface operation, or <code>null</code> if undefined.
     */
    public SOAPBindingOperation(QName name, SOAPBindingMessageReference inputMessageReference,
            SOAPBindingMessageReference outputMessageReference,
            Map<QName, SOAPBindingFaultReference> faultReferences, ElementPolicy elementPolicy,
            SOAPMessageExchangePattern soapMessageExchangePattern, String soapAction) {
        super(name, inputMessageReference, outputMessageReference, faultReferences, elementPolicy);
        this.soapMessageExchangePattern = soapMessageExchangePattern;
        this.soapAction = soapAction;
    }

    /**
     * Identifies value of SOAP Action for the initial message of the message exchange pattern
     * of the bound interface operation.
     * 
     * @return String soap action or <code>null</code> if action is undefined.
     */
    public String getSoapAction() {
        return soapAction;
    }

    /**
     * Identifies protocol specific, i.e. SOAP, message exchange pattern utilized by this
     * operation.
     * <p>
     * If {@link SOAPVersion} employed by parent <code>Binding</code> is soap 1.1, no
     * assertion is made regarding soapMessageExchangePattern, i.e. argument MUST be null.
     * 
     * @return SOAPMessageExchangePattern or null if no assertion made regarding soap MEP.
     */
    public SOAPMessageExchangePattern getSoapMessageExchangePattern() {
        return soapMessageExchangePattern;
    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("name: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(name);
        sb.append(System.getProperty("line.separator"));
        sb.append("soapAction: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(soapAction);
        sb.append(System.getProperty("line.separator"));
        sb.append("soapMessageExchangePattern: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(soapMessageExchangePattern);
        sb.append(System.getProperty("line.separator"));
        sb.append("inputMessageReference: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(inputMessageReference);
        sb.append(System.getProperty("line.separator"));
        sb.append("outputMessageReference: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(outputMessageReference);
        sb.append(System.getProperty("line.separator"));
        sb.append("faultReferences: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(faultReferences);
        return sb.toString();
    }

    @Override
    public int hashCode() {
        return hashCode(name)
                + hashCode(soapAction)
                + hashCode(soapMessageExchangePattern)
                + hashCode(inputMessageReference)
                + hashCode(outputMessageReference)
                + hashCode(faultReferences);
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof SOAPBindingOperation)) {
            return false;
        }

        if ((name == ((SOAPBindingOperation)obj).name || (name != null && name
                .equals(((SOAPBindingOperation)obj).name)))
                && (soapAction == ((SOAPBindingOperation)obj).soapAction || (soapAction != null && soapAction
                        .equals(((SOAPBindingOperation)obj).soapAction)))
                && (soapMessageExchangePattern == ((SOAPBindingOperation)obj).soapMessageExchangePattern || (soapMessageExchangePattern != null && soapMessageExchangePattern
                        .equals(((SOAPBindingOperation)obj).soapMessageExchangePattern)))
                && (inputMessageReference == ((SOAPBindingOperation)obj).inputMessageReference || (inputMessageReference != null && inputMessageReference
                        .equals(((SOAPBindingOperation)obj).inputMessageReference)))
                && (outputMessageReference == ((SOAPBindingOperation)obj).outputMessageReference || (outputMessageReference != null && outputMessageReference
                        .equals(((SOAPBindingOperation)obj).outputMessageReference)))
                && (faultReferences == ((SOAPBindingOperation)obj).faultReferences || (faultReferences != null && faultReferences
                        .equals(((SOAPBindingOperation)obj).faultReferences)))) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}
