/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.binding.soap.rt;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import javax.xml.XMLConstants;
import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.binding.rt.BindingFaultRT;
import org.bluestemsoftware.specification.eoa.component.message.rt.Content;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPVersion;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

/**
 * A version agnostic fault generated by a SOAPBinding and/or BindingModule which processes
 * SOAP messages.
 */
public abstract class SOAPFault extends BindingFaultRT {

    private static final long serialVersionUID = 1L;

    public static final String DEFAULT_ACTION = "http://www.w3.org/2005/08/addressing/soap/fault";
    public static final String SOAP_NS_PREFIX = "soap";

    private static final QName HEADER_NAME = new QName("*", "Header"); // version agnostic
    private static final QName BODY_NAME = new QName("*", "Body"); // version agnostic
    private static final QName FAULT_NAME = new QName("*", "Fault"); // version agnostic

    private static final QName SOAP_11_REASON_NAME = new QName("faultstring");
    private static final QName SOAP_12_REASON_NAME = new QName(SOAPVersion.SOAP_12_ENVELOPE_NS, "Reason");
    private static final QName SOAP_12_TEXT_NAME = new QName(SOAPVersion.SOAP_12_ENVELOPE_NS, "Text");
    
    private static final QName SOAP_11_DETAIL_NAME = new QName("detail");
    private static final QName SOAP_12_DETAIL_NAME = new QName(SOAPVersion.SOAP_12_ENVELOPE_NS, "Detail");

    private static final QName SOAP_11_CODE_NAME = new QName("faultcode");
    private static final QName SOAP_12_CODE_NAME = new QName(SOAPVersion.SOAP_12_ENVELOPE_NS, "Code");
    private static final QName SOAP_12_VALUE_NAME = new QName(SOAPVersion.SOAP_12_ENVELOPE_NS, "Value");
    private static final QName SOAP_12_SUBCODE_NAME = new QName(SOAPVersion.SOAP_12_ENVELOPE_NS, "Subcode");

    protected SOAPVersion soapVersion;

    public SOAPFault(SOAPVersion soapVersion, QName faultName, String faultReason) {
        super(faultName, faultReason);
        if (soapVersion == null) {
            throw new IllegalArgumentException("soapVersion null");
        } else {
            this.soapVersion = soapVersion;
        }
    }

    public SOAPFault(SOAPVersion soapVersion, QName faultName, String faultReason, Throwable th) {
        super(faultReason, th);
        if (soapVersion == null) {
            throw new IllegalArgumentException("soapVersion null");
        } else {
            this.soapVersion = soapVersion;
        }
    }

    /**
     * Defines SOAP version which MUST be used to set namespace of {@link #getFaultCode()}.
     * 
     * @return
     */
    public SOAPVersion getSoapVersion() {
        return soapVersion;
    }

    /**
     * Returns a version specific 'top-level' fault code, i.e. which gives a 'general'
     * indication of fault, e.g. env:Sender, env:VersionMismatch, etc ...
     * 
     * @return
     */
    public abstract QName getFaultCode();

    /**
     * Returns sub-codes which give a more specific indication of fault within context of fault
     * code. Note that the MOST specific indication of fault SHOULD be the last entry in list.
     * <p>
     * Note that the default implementation returns a singleton list containing faultName if
     * defined, otherwise an empty list is returned. Method SHOULD be overridden by concrete
     * instances of this class to provide a more specific indication of fault.
     * 
     * @return subcodes or an empty list if none defined.
     */
    @SuppressWarnings("unchecked")
    public List<QName> getFaultSubCodes() {
        if (faultName != null) {
            return Collections.singletonList(faultName);
        } else {
            return Collections.EMPTY_LIST;
        }
    }

    /**
     * Returns default soap fault action defined by ws-addressing specification. Method SHOULD
     * be overridden by concrete instances of this class to provide a more specific action.
     */
    @Override
    public String getAction() {
        return DEFAULT_ACTION;
    }

    /**
     * Retrieves soap fault element from a version specific fault envelope.
     * 
     * @param faultEnvelope
     * @return fault element, or <code>null</code> if not found.
     */
    public static Element getFaultElement(Element faultEnvelope) {
        Element bodyElement = getChildElement(faultEnvelope, BODY_NAME);
        if (bodyElement == null) {
            return null;
        }
        return getChildElement(bodyElement, FAULT_NAME);
    }

    /**
     * Retrieves fault detail element from a version specific fault envelope.
     * 
     * @param faultEnvelope
     * @return detail element, or <code>null</code> if not found.
     */
    public static Element getFaultDetailElement(Element faultEnvelope) {
        Element faultElement = SOAPFault.getFaultElement(faultEnvelope);
        if (faultElement == null) {
            return null;
        }
        SOAPVersion sv = SOAPVersion.valueOfEnvelopeNamespace(faultEnvelope.getNamespaceURI());
        if (sv == SOAPVersion.SOAP11) {
            return getChildElement(faultElement, SOAP_11_DETAIL_NAME);
        } else {
            return getChildElement(faultElement, SOAP_12_DETAIL_NAME);
        }
    }
    
    /**
     * Retrieves fault detail, i.e. content contained within fault detail element, from a version specific fault envelope.
     * 
     * @param faultEnvelope
     * @return fault detail, or empty list if none found.
     */
    public static List<Content> getFaultDetail(Element faultEnvelope) {
        Element faultDetailElement = SOAPFault.getFaultDetailElement(faultEnvelope);
        if (faultDetailElement == null) {
            return null;
        }
        return getChildElements(faultDetailElement);
    }

    /**
     * Retrieves header element from a version specific fault envelope.
     * 
     * @param faultEnvelope
     * @return header element, or <code>null</code> if not found.
     */
    public static Element getFaultHeaderElement(Element faultEnvelope) {
        return getChildElement(faultEnvelope, HEADER_NAME);
    }
    
    /**
     * Extracts the fault reason from a version specific fault envelope.
     * 
     * @param faultEnvelope
     * @return QName fault name
     */
    public static String getFaultReason(Element faultEnvelope) {
        Element faultElement = SOAPFault.getFaultElement(faultEnvelope);
        if (faultElement == null) {
            return null;
        }
        SOAPVersion sv = SOAPVersion.valueOfEnvelopeNamespace(faultEnvelope.getNamespaceURI());
        if (sv == SOAPVersion.SOAP11) {
            Element faultStringElement = getChildElement(faultElement, SOAP_11_REASON_NAME);
            if (faultStringElement == null) {
                return null;
            }
            return getText(faultStringElement);
        } else {
            Element faultReasonElement = getChildElement(faultElement, SOAP_12_REASON_NAME);
            if (faultReasonElement == null) {
                return null;
            }
            Element textElement = getChildElement(faultReasonElement, SOAP_12_TEXT_NAME);
            if (textElement == null) {
                return null;
            }
            return getText(textElement);
        }
    }

    /**
     * Extracts the most specific indication of fault from a version specific fault envelope.
     * 
     * @param faultEnvelope
     * @return QName fault name
     */
    public static QName getFaultName(Element faultEnvelope) {
        Element faultElement = SOAPFault.getFaultElement(faultEnvelope);
        if (faultElement == null) {
            return null;
        }
        SOAPVersion sv = SOAPVersion.valueOfEnvelopeNamespace(faultEnvelope.getNamespaceURI());
        if (sv == SOAPVersion.SOAP11) {
            Element codeElement = getChildElement(faultElement, SOAP_11_CODE_NAME);
            if (codeElement == null) {
                return null;
            }
            String prefixedName = getText(codeElement);
            if (prefixedName != null) {
                return ComponentName.valueOf(codeElement, prefixedName);
            } else {
                return null;
            }
        } else {
            Element codeElement = getChildElement(faultElement, SOAP_12_CODE_NAME);
            if (codeElement == null) {
                return null;
            }
            Element subcodeElement = getChildElement(codeElement, SOAP_12_SUBCODE_NAME);
            if (subcodeElement == null) {
                Element valueElement = getChildElement(codeElement, SOAP_12_VALUE_NAME);
                if (valueElement != null) {
                    String prefixedName = getText(valueElement);
                    return ComponentName.valueOf(valueElement, prefixedName);
                } else {
                    return null;
                }                
            }
            QName result = null;
            Element nextSubcodeElement = subcodeElement;
            while (nextSubcodeElement != null) {
                Element valueElement = getChildElement(nextSubcodeElement, SOAP_12_VALUE_NAME);
                if (valueElement != null) {
                    String prefixedName = getText(valueElement);
                    result = ComponentName.valueOf(valueElement, prefixedName);
                } 
                nextSubcodeElement= getNextSubcodeElement(nextSubcodeElement);
            }
            return result;  
        }
    }

    private static Element getNextSubcodeElement(Element subcodeElement) {
        return getChildElement(subcodeElement, SOAP_12_SUBCODE_NAME);
    }

    /**
     * Returns first encountered element with given qualified name within context of parent
     * element.
     * @param parent
     * @param childName
     *        qualified name of child. note that wild card '*' may be set as value for
     *        namespaceURI or localName. For an element with no namespace, leave ns uri to null.
     * @return
     */
    protected static Element getChildElement(Element parent, QName childName) {
        NodeList nodeList = parent.getChildNodes();
        for (int i = 0; i < nodeList.getLength(); i++) {
            if (nodeList.item(i).getNodeType() != Node.ELEMENT_NODE) {
                continue;
            }
            Element element = (Element)nodeList.item(i);
            if (childName.getNamespaceURI().equals(XMLConstants.NULL_NS_URI)) {
                if (element.getNamespaceURI() != null) {
                    continue;
                }
            } else {
                if (element.getNamespaceURI() == null) {
                    continue;
                }
                if (!childName.getNamespaceURI().equals("*")
                        && !element.getNamespaceURI().equals(childName.getNamespaceURI())) {
                    continue;
                }
            }
            if (!childName.getLocalPart().equals("*") && !element.getLocalName().equals(childName.getLocalPart())) {
                continue;
            }
            return element;
        }
        return null;
    }
    
    private static List<Content> getChildElements(Element parent) {
        List<Content> result = new ArrayList<Content>();
        NodeList nodeList = parent.getChildNodes();
        for (int i = 0; i < nodeList.getLength(); i++) {
            if (nodeList.item(i).getNodeType() != Node.ELEMENT_NODE) {
                continue;
            }
            result.add((Content)nodeList.item(i));
        }
        return result;
    }

    private static String getText(Element element) {
        if (element == null || !element.hasChildNodes()) {
            return null;
        }
        try {
            Text textNode = (Text)element.getFirstChild();
            return textNode.getData();
        } catch (ClassCastException ce) {
        }
        return null;
    }

}
