/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.binding.soap.rt;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.component.RootComponent;
import org.bluestemsoftware.specification.eoa.component.message.rt.Content;
import org.bluestemsoftware.specification.eoa.component.message.rt.ContentFactory;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPVersion;
import org.w3c.dom.Element;

/**
 * A 'catch-all' SOAPFault indicating fault is due to 'reasons attributable to the processing
 * of the message rather than to the contents of the message itself.'
 */

public class SOAPReceiverFault extends SOAPFault {

    private static final long serialVersionUID = 1L;

    public static final String DEFAULT_REASON = "An error occurred while processing message";

    public static final QName SOAP_11_FAULT_CODE = new QName("http://schemas.xmlsoap.org/soap/envelope/",
            "Server", SOAP_NS_PREFIX);
    public static final QName SOAP_12_FAULT_CODE = new QName("http://www.w3.org/2003/05/soap-envelope",
            "Receiver", SOAP_NS_PREFIX);

    /**
     * Constructs a <code>SOAPReceiverFault</code>.
     * 
     * @param soapVersion
     * @param faultReason
     */
    public SOAPReceiverFault(SOAPVersion soapVersion, String faultReason) {
        super(soapVersion, RECEIVER_ERROR, faultReason);
    }

    /**
     * Constructs a <code>SOAPReceiverFault</code> .
     * 
     * @param soapVersion
     * @param faultReason
     * @param th
     *        Cause of fault. Note that stack trace will be embedded within fault detail. Use
     *        alternative constructor {@link #SOAPReceiverFault(SOAPVersion, String)} if this
     *        behavior is not desired.
     */
    public SOAPReceiverFault(SOAPVersion soapVersion, String faultReason, Throwable th) {
        super(soapVersion, RECEIVER_ERROR, faultReason, th);
    }

    /**
     * Constructor used by subclasses which provide a more specific indication of fault.
     * 
     * @param soapVersion
     * @param faultName
     * @param faultReason
     */
    protected SOAPReceiverFault(SOAPVersion soapVersion, QName faultName, String faultReason) {
        super(soapVersion, faultName, faultReason);
    }

    /**
     * Constructs a <code>SOAPReceiverFault</code> with implied reason
     * {@link #DEFAULT_REASON} and a provided cause.
     * 
     * @param th
     *        Cause of fault. Note that stack trace will be embedded within fault detail. Use
     *        alternative constructor {@link #SOAPReceiverFault(SOAPVersion, String)} if this
     *        behavior is not wanted.
     */
    public SOAPReceiverFault(SOAPVersion soapVersion, Throwable th) {
        super(soapVersion, RECEIVER_ERROR, DEFAULT_REASON, th);
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPFault#getFaultCode(org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPVersion)
     */
    @Override
    public QName getFaultCode() {
        if (soapVersion == SOAPVersion.SOAP11) {
            int size = getFaultSubCodes().size();
            if (size > 0) {
                return getFaultSubCodes().get(size - 1);
            } else if (faultName != null) {
                return faultName;
            } else {
                return SOAP_11_FAULT_CODE;
            }
        } else {
            return SOAP_12_FAULT_CODE;
        }
    }
    
    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.binding.rt.BindingFaultRT#appendFaultDetail(org.bluestemsoftware.specification.eoa.component.message.rt.Content)
     */
    @Override
    public void appendFaultDetail(Content faultEnvelope) {
        
        // if cause of fault was due to throwable, we'll embed the stack trace
        // within detail element. note that binding controls this behavior, i.e.
        // if user preference is that stack trace should be suppressed, binding
        // should construct fault without passing a throwable
        
        if (getCause() == null) {
            return;
        }
        
        Element faultDetail = getFaultDetailElement(faultEnvelope);
        ContentFactory factory = faultEnvelope.getFactory();
        Element ste = factory.createElementNS(RootComponent.EOA_COMPONENT_NS, "eoa:StackTrace");
        faultDetail.appendChild(ste);
        StringBuilder sb = new StringBuilder("<![CDATA[");
        sb.append(getCause().toString());
        sb.append("]]>");
        ste.appendChild(factory.createTextNode(sb.toString()));
        
    }
    
}
