/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.binding.soap.rt;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.component.message.rt.Content;
import org.bluestemsoftware.specification.eoa.component.message.rt.ContentFactory;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPVersion;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * A non-specific SOAPFault which indicates fault is due to a problem with message sent by
 * sender, i.e. it is 'incorrectly formed or did not contain the appropriate information in
 * order to succeed'.
 */
public class SOAPSenderFault extends SOAPFault {

    private static final long serialVersionUID = 1L;

    public static final QName SOAP_11_FAULT_CODE = new QName("http://schemas.xmlsoap.org/soap/envelope/",
            "Client", SOAP_NS_PREFIX);
    public static final QName SOAP_12_FAULT_CODE = new QName("http://www.w3.org/2003/05/soap-envelope", "Sender",
            SOAP_NS_PREFIX);

    private Content payload;

    protected SOAPSenderFault(SOAPVersion soapVersion, QName faultName, String faultReason) {
        super(soapVersion, faultName, faultReason);
    }

    /**
     * 
     * @param soapVersion
     * @param faultReason
     */
    public SOAPSenderFault(SOAPVersion soapVersion, String faultReason) {
        super(soapVersion, SENDER_ERROR, faultReason);
    }

    /**
     * Allows thrower to pass payload of offending message which will be embedded within fault detail
     * element of fault envelope generated from this fault.
     * 
     * @param soapVersion
     * @param faultReason
     * @param payload
     */
    public SOAPSenderFault(SOAPVersion soapVersion, String faultReason, Content payload) {
        super(soapVersion, SENDER_ERROR, faultReason);
        this.payload = payload;
    }
    
    /**
     * Gets payload of offending message.
     * 
     * @return payload or <code>null</code> if undefined.
     */
    public Content getPayload() {
        return payload;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPFault#getFaultCode(org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPVersion)
     */
    @Override
    public QName getFaultCode() {
        if (soapVersion == SOAPVersion.SOAP11) {
            int size = getFaultSubCodes().size();
            if (size > 0) {
                return getFaultSubCodes().get(size - 1);
            } else if (faultName != null) {
                return faultName;
            } else {
                return SOAP_11_FAULT_CODE;
            }
        } else {
            return SOAP_12_FAULT_CODE;
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.binding.rt.BindingFaultRT#appendFaultDetail(org.bluestemsoftware.specification.eoa.component.message.rt.Content)
     */
    @Override
    public void appendFaultDetail(Content faultEnvelope) {
        if (payload == null) {
            return;
        }
        ContentFactory factory = faultEnvelope.getFactory();
        Node imported = null;
        try {
            imported = factory.importContent(payload);
        } catch (UnsupportedOperationException ue) {
            imported = factory.importNode(payload, true);
        }
        Element faultDetailElement = getFaultDetailElement(faultEnvelope);
        faultDetailElement.appendChild(imported);
    }

}
