/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.feature.http.client;

import org.bluestemsoftware.specification.eoa.ext.feature.auth.HostInfo;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.HTTPAuthenticationScheme;
import org.bluestemsoftware.specification.eoa.system.server.Feature;

/**
 * A factory for creating configurable {@link HTTPClient} objects which can be used to connect
 * to host(s) and send requests using HTTP transmission protocol.
 */
public class HTTPClientFeature extends Feature {

    public static final String TYPE = HTTPClientFeature.class.getName();

    /*
     * (non-Javadoc)
     * 
     * SPI
     */
    public interface Provider extends Feature.Provider {

        public HTTPClient spi_createHTTPClient() throws HTTPClientFeatureException;

        public HTTPClient spi_createHTTPClient(HostInfo hostInfo, ClientAuthInfo authInfo) throws HTTPClientFeatureException;

    }

    public HTTPClientFeature(HTTPClientFeatureFactory factory, Feature.Provider provider) {
        super(factory, provider);
    }

    /**
     * Creates a non-authenticating, host agnostic <code>HTTPClient</code> instance, i.e.
     * which may be used to send messages to any unsecured host.
     * 
     * @return HTTPClient
     * @throws HTTPClientFeatureException
     */
    public HTTPClient createHTTPClient() throws HTTPClientFeatureException {
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            return ((Provider)provider).spi_createHTTPClient();
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /**
     * Creates a host specific <code>HTTPClient</code> instance, i.e. a client which connects
     * to indicated host[:port] and which uses host specific configuration which SHOULD be
     * specified within feature configuration.
     * 
     * @param hostInfo
     *        identifies host/port
     * @param authInfo
     *        or <tt>null</tt> if no authentication required.
     * 
     * @return HTTPClient
     * @throws HTTPClientFeatureException
     */
    public HTTPClient createHTTPClient(HostInfo hostInfo, ClientAuthInfo authInfo) throws HTTPClientFeatureException {
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            return ((Provider)provider).spi_createHTTPClient(hostInfo, authInfo);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /**
     * Encapsulates data used by client to authenticate user in response to challenges from
     * host.
     */
    public interface ClientAuthInfo {

        /**
         * Gets password.
         * @return
         */
        public String getPassword();

        /**
         * Gest realm.
         * @return realm or <code>null</code> if credentials are to be sent in response to
         *         any realm challenge.
         */
        public String getRealm();

        /**
         * Gest scheme.
         * @return scheme or <code>null</code> if credentials are to be sent in response to
         *         any scheme challenge.
         */
        public HTTPAuthenticationScheme getScheme();

        /**
         * Gets user name.
         * @return user name.
         */
        public String getUserName();

    }

}
