/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http;

import java.util.Collection;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction.Direction;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.TransportProtocol;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.http.rt.HyperTextTransferProtocolRT;

/**
 * Statically describes the 'Hyper Text Transfer Protocol' which affects runtime transport
 * behavior as modeled by {@link HyperTextTransferProtocolRT}.
 */
public interface HyperTextTransferProtocol extends TransportProtocol {

    public static final String HOST_VARIABLE = "${host}";
    public static final String PORT_VARIABLE = "${port}";
    public static final String SSL_PORT_VARIABLE = "${ssl.port}";

    public static String TYPE = HyperTextTransferProtocol.class.getName();

    /**
     * Generates an HTTP <code>Endpoint</code> address using indicated path, query and
     * fragment. Compliant implementations MUST use use the following values to construct
     * address:
     * <p>
     * scheme is 'http'
     * <p>
     * host is machine's host name
     * <p>
     * port is {@link HTTPServerFeature#getPort(false)}
     * <p>
     * path is as indicated
     * <p>
     * query is as indicated
     * <p>
     * fragment is as indicated
     * 
     * @param path
     *        path used to construct endpoint address
     * @param query
     *        query used to construct endpoint address or <code>null</code> if undefined
     * @param fragment
     *        fragment used to construct endpoint address or <code>null</code> if undefined
     * @return generated address
     */
    public String createEndpointAddress(String path, String query, String fragment) throws DeploymentException;

    /**
     * Indicates the default value for the HTTP Request Method for all the Interface Operation
     * components of any Interface component to which Binding is applied.
     * @return
     */
    public HTTPMethod getMethodDefault();

    /**
     * Indicates whether or not HTTP cookies are used on some or on all operations of parent
     * <code>Binding</code> as defined by IETF RFC 2965.
     * 
     * @return True if used. False if not, Null if no assertion is made regarding cookies.
     */
    public Boolean useHTTPCookies();

    /**
     * Returns HTTP headers defined for a specific fault.
     * @param faultName
     * @return Collection of headers or empty collection if none defined.
     */
    public Collection<HTTPHeader> getFaultHTTPHeaders(QName faultName);

    /**
     * Returns HTTP headers defined for normal message with indicated direction on indicated
     * operation.
     * 
     * @param operationName
     * @param direction
     * @return Collection of headers or empty collection if none defined.
     */
    public Collection<HTTPHeader> getMessageReferenceHTTPHeaders(QName operationName, Direction direction);

    /**
     * Defines content encodings available for indicated fault.
     * 
     * @param faultName
     * @return String a comma delimited list of values or <code>null</code> if undefined.
     */
    public String getFaultHTTPContentEncoding(QName faultName);

    /**
     * Identifies the error Status-Code as defined by [IETF RFC 2616] that the service will use
     * in case the fault is returned.
     * 
     * @param faultName
     * @return status code or <code>null</code> if no claim is made by service.
     * 
     * @see http://www.ietf.org/rfc/rfc2616.txt
     */
    public Integer getErrorStatusCode(QName faultName);
    
    /**
     * Gets content encodings available for a normal message with indicated direction on
     * indicated operation.
     * 
     * @param operationName
     * @param direction
     * @return String a comma delimited list of values or <code>null</code> if undefined.
     */
    public String getMessageReferenceHTTPContentEncoding(QName operationName, Direction direction);

    /**
     * Gets query parameter separator character for indicated operation.
     * 
     * @param operationName
     * @return String or <code>null</code> if undefined.
     */
    public String getHTTPQueryParameterSeparator(QName operationName);

    /**
     * Gets location info object for indicated operation.
     * @param operationName
     * @return HTTPLocationInfo or <code>null</code> if undefined.
     */
    public HTTPLocationInfo getHTTPLocationInfo(QName operationName);

    /**
     * Gets location info object for indicated location.
     * @param location
     *        which MUST be unique within context of binding.
     * @return
     */
    public HTTPLocationInfo getHTTPLocationInfo(String location);

    /**
     * Gets <code>HTTPMethod</code> for indicated operation.
     * @param operationName
     * @return HTTPMethod or <code>null</code> if undefined.
     */
    public HTTPMethod getMethod(QName operationName);

    /**
     * Gets serialization supported for input message of indicated operation.
     * @param operationName
     * @return String formatted according to rules defined within WSDL 2.0 adjunct spec or
     *         <code>null</code> if undefined.
     * @see http://www.w3.org/TR/wsdl20-adjuncts/#_http_ser_val
     */
    public String getInputSerialization(QName operationName);

    /**
     * Gets serialization supported for output message of indicated operation.
     * @param operationName
     * @return String formatted according to rules defined within WSDL 2.0 adjunct spec or
     *         <code>null</code> if undefined.
     * @see http://www.w3.org/TR/wsdl20-adjuncts/#_http_ser_val
     */
    public String getOutputSerialization(QName operationName);

    /**
     * Gets serialization supported for fault messages of indicated operation.
     * @param operationName
     * @return String formatted according to rules defined within WSDL 2.0 adjunct spec or
     *         <code>null</code> if undefined.
     * @see http://www.w3.org/TR/wsdl20-adjuncts/#_http_ser_val
     */
    public String getFaultSerialization(QName operationName);

}
