/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.idl.wsdl;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.ext.idl.IDLFactory;

public final class WSDLLocation {

    private String scheme;
    private String host;
    private int port = -1;
    private String path;
    private String versionLabel;
    private String targetNamespace;
    private Set<FragmentIdentifier> fragmentIdentifiers = new HashSet<FragmentIdentifier>();
    private static Pattern pathExpression = Pattern.compile(".+[/]eoa[/]wsdl[/](\\d{2,})");

    /**
     * Constructs a <code>WSDLLocation</code> object. Used by concrete implementations of
     * <code>WSDLFactory</code> to construct a location URI for wsdl include\import elements
     * when writing a component(s) definition.
     * 
     * @param wsdlLocation
     *        location of base document from which scheme, authority and path will be used to
     *        construct this instance
     * 
     * @return WSDLLocation
     */
    public WSDLLocation(WSDLLocation wsdlLocation) {
        if (wsdlLocation == null) {
            throw new IllegalArgumentException("wsdlLocation null");
        }
        this.scheme = wsdlLocation.getScheme();
        this.host = wsdlLocation.getHost();
        this.port = wsdlLocation.getPort();
        this.path = wsdlLocation.getPath();
    }

    private WSDLLocation(String scheme, String host, int port, String path) {
        this.scheme = scheme;
        this.host = host;
        this.port = port;
        this.path = path;
    }

    /**
     * Adds a component which MUST be described by wsdl document dereferenced from this
     * location. Note that if more than one fragmentID is defined, they MUST share a common
     * namespace
     * 
     * @param fragmentIdentifier
     */
    public void addFragmentIdentifier(FragmentIdentifier fragmentIdentifier) {
        if (fragmentIdentifier == null) {
            throw new IllegalArgumentException("fragmentIdentifier null");
        }
        if (targetNamespace == null) {
            targetNamespace = fragmentIdentifier.getRootComponentNamespace();
        } else {
            if (!targetNamespace.equals(fragmentIdentifier.getRootComponentNamespace())) {
                throw new IllegalArgumentException("FragmentIdentifiers MUST all share a common namespace.");
            }
        }
        fragmentIdentifiers.add(fragmentIdentifier);
    }

    public String getHost() {
        return host;
    }

    public int getPort() {
        return port;
    }

    public String getScheme() {
        return scheme;
    }

    public String getPath() {
        return path;
    }

    /**
     * Gets identifiers of components to describe all of which MUST be defined within the same
     * namespace
     * 
     * @return identifiers or an empty set if none defined, i.e. which implies that all
     *         components within target namespace should be described.
     */
    public Set<FragmentIdentifier> getFragmentIdentifiers() {
        return fragmentIdentifiers;
    }

    /**
     * Parses value returned via {@link #getPath()} and returns version label which identifies
     * targeted wsdl version of imported\included definition, i.e. 11, 20.
     * 
     * @return version label or empty string if path fails to match the following regular
     *         expression: '.+[/]eoa[/]wsdl[/](\\d{2,})'.
     * @see IDLFactory#getVersionLabel()
     */
    public String getVersionLabel() {
        if (versionLabel == null) {
            Matcher pathMatcher = pathExpression.matcher(path);
            if (pathMatcher.find()) {
                versionLabel = pathMatcher.group(1);
            } else {
                path = "";
            }
        }
        return versionLabel;
    }

    /**
     * Parses a request URI into a WSDLLocation instance.
     * 
     * @param locationURI
     * @return
     */
    public static WSDLLocation valueOf(String locationURI) {
        URI temp = null;
        try {
            temp = new URI(locationURI);
        } catch (URISyntaxException ue) {
            throw new IllegalArgumentException("Invalid locationURI. " + ue);
        }
        WSDLLocation wsdlLocation = new WSDLLocation(temp.getScheme(), temp.getHost(), temp.getPort(), temp
                .getPath());
        if (temp.getFragment() != null) {
            for (String expression : temp.getFragment().split(",")) {
                wsdlLocation.addFragmentIdentifier(FragmentIdentifier.valueOf(expression));
            }
        }
        return wsdlLocation;
    }

    /**
     * Converts this <code>WSDLLocation</code> instance into a URI by constructing a URI with
     * the following arguments where ${foo} represents value of instance variable foo:
     * <p>
     * scheme = ${scheme}
     * <p>
     * host = ${host}
     * <p>
     * port = ${port}
     * <p>
     * userInfo = <code>null</code>
     * <p>
     * path = ${path}
     * <p>
     * query = <code>null</code>
     * <p>
     * fragment = [${fragmentIdentifier}[,${fragmentIdentifier}]]
     */
    public URI toURI() {

        StringBuilder fragment = null;
        if (fragmentIdentifiers.size() > 0) {
            List<String> sortedList = new ArrayList<String>();
            for (FragmentIdentifier fid : fragmentIdentifiers) {
                sortedList.add(fid.toString());
            }
            Collections.sort(sortedList);
            fragment = new StringBuilder();
            for (String fid : sortedList) {
                fragment.append(fragment.length() == 0 ? fid : "," + fid);
            }
        }

        URI answer = null;
        try {
            if (fragment != null) {
                answer = new URI(scheme, null, host, port, path, null, fragment.toString());
            } else {
                answer = new URI(scheme, null, host, port, path, null, null);
            }
        } catch (URISyntaxException fatchance) {
        }

        return answer;

    }

    @Override
    public String toString() {
        return toURI().toString();
    }

    @Override
    public final int hashCode() {
        return hashCode(this.toString());
    }

    @Override
    public final boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }
        if (!(obj instanceof WSDLLocation)) {
            return false;
        }
        return ((WSDLLocation)obj).toURI().equals(this.toURI());
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}
