/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the Ext EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 * 
 * [1] http://alakai.org/licensing/specification-eoa-ext-license.html
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl11.ext;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.component.binding.BindingFaultReference;
import org.bluestemsoftware.specification.eoa.component.binding.BindingMessageReference;
import org.bluestemsoftware.specification.eoa.component.binding.BindingOperation;
import org.bluestemsoftware.specification.eoa.component.intrface.Interface;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceOperation;
import org.bluestemsoftware.specification.eoa.component.policy.rt.PolicyExpression;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPBinding;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPBindingMessageReference;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPBindingOperation;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPDocumentStyleMessage;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPHeaderBlock;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPMessage;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPMessageStyle;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPRpcStyleMessage;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPVersion;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.WSDLException;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl11.BindingElement;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl11.BindingOperationElement;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl11.BindingOperationFaultElement;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl11.BindingOperationInputElement;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl11.BindingOperationOutputElement;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl11.DefinitionsElement;
import org.bluestemsoftware.specification.eoa.ext.idl.wsdl.wsdl11.WSDL11Document;
import org.bluestemsoftware.specification.eoa.ext.policy.PolicyException;
import org.w3c.dom.Element;

public class SOAPBindingWriter {

    public static void writeBinding(WSDL11Document wsdlDocument, SOAPBinding binding) throws WSDLException {

        SOAPVersion soapVersion = binding.getSoapVersion();
        
        DefinitionsElement def = wsdlDocument.getDefinitionsElement();

        QName bindingName = binding.getName();
        BindingElement bindingElement = wsdlDocument.createBindingElement();
        bindingElement.setName(bindingName.getLocalPart());
        if (binding.getInterfaceReference() != null) {
            bindingElement.setType(binding.getInterfaceReference().getReferencedComponentName());
        }
        def.addBindingElement(bindingElement);

        if (binding.getElementPolicy() != null) {
            PolicyExpression policy;
            try {
                policy = binding.getElementPolicy().getMergedExpression();
            } catch (PolicyException pe) {
                throw new WSDLException("Error writing definition. " + pe);
            }
            writePolicy(policy, bindingElement);
        }

        // derive binding style by examining style of a normal, i.e. non-fault
        // message, all of which use a consistent style as enforced by basic
        // profile

        String bindingStyle = "document";
        for (BindingOperation bindingOperation : binding.getOperations()) {
            BindingMessageReference inputMessageRef = bindingOperation.getInputMessageReference();
            if (inputMessageRef != null) {
                SOAPMessage sm = ((SOAPBindingMessageReference)inputMessageRef).getSoapMessage();
                if (sm.getMessageStyle() == SOAPMessageStyle.RPC_STYLE) {
                    bindingStyle = "rpc";
                    break;
                }
            }
        }

        // create the soap:binding element
        SOAPBindingElement soapBindingElement = null;
        if (soapVersion == SOAPVersion.SOAP11) {
            soapBindingElement= wsdlDocument.createSOAP11BindingElement();
        } else {
            soapBindingElement= wsdlDocument.createSOAP12BindingElement();
        }
        
        soapBindingElement.setStyle(bindingStyle);
        soapBindingElement.setTransport("http://schemas.xmlsoap.org/soap/http");
        bindingElement.setSOAPBindingElement(soapBindingElement);

        // retrieve interface, which must not be null, i.e. if binding is reusable,
        // it cannot be described via wsdl 11

        Interface intrface = null;
        if (binding.getInterfaceReference() == null) {
            throw new WSDLException("Error writing binding "
                    + bindingName
                    + ". Binding must reference an interface.");
        } else {
            intrface = binding.getInterfaceReference().getReferencedComponent();
        }

        // create the binding operations
        for (BindingOperation bo : binding.getOperations()) {
            SOAPBindingOperation sbo = (SOAPBindingOperation)bo;
            InterfaceOperation io = intrface.getOperation(bo.getName());
            writeBindingOperation(wsdlDocument, soapVersion, sbo, io, bindingElement, bindingStyle);
        }

    }

    private static void writeBindingOperation(WSDL11Document wsdlDocument, SOAPVersion soapVersion, SOAPBindingOperation bindingOperation, InterfaceOperation interfaceOperation, BindingElement bindingElement, String bindingStyle) throws WSDLException {

        // create the operation element
        BindingOperationElement bindingOperationElement = wsdlDocument.createBindingOperationElement();
        bindingOperationElement.setName(bindingOperation.getName().getLocalPart());
        bindingElement.addOperationElement(bindingOperationElement);

        if (bindingOperation.getElementPolicy() != null) {
            PolicyExpression policy;
            try {
                policy = bindingOperation.getElementPolicy().getMergedExpression();
            } catch (PolicyException pe) {
                throw new WSDLException("Error writing definition. " + pe);
            }
            writePolicy(policy, bindingOperationElement);
        }

        // create the soap:operation element
        SOAPOperationElement soapOperation = null;
        if (soapVersion == SOAPVersion.SOAP11) {
            soapOperation = wsdlDocument.createSOAP11OperationElement();
        } else {
            soapOperation= wsdlDocument.createSOAP12OperationElement();
        }
        
        if (bindingOperation.getSoapAction() != null) {
            soapOperation.setSOAPAction(bindingOperation.getSoapAction());
        }
        soapOperation.setStyle(bindingStyle);
        bindingOperationElement.setSOAPOperationElement(soapOperation);

        // create the input element and attach soap:body and optional headers
        BindingOperationInputElement bindingInputElement = wsdlDocument.createBindingOperationInputElement();
        BindingMessageReference inputMessageReference = bindingOperation.getInputMessageReference();

        if (inputMessageReference.getElementPolicy() != null) {
            PolicyExpression policy;
            try {
                policy = inputMessageReference.getElementPolicy().getMergedExpression();
            } catch (PolicyException pe) {
                throw new WSDLException("Error writing definition. " + pe);
            }
            writePolicy(policy, bindingInputElement);
        }

        // process the body first. if this is a doc-lit style binding,
        // body contains zero or one part. if it's rpc-lit, it's a comma
        // separated list

        SOAPMessage soapMessage = ((SOAPBindingMessageReference)inputMessageReference).getSoapMessage();

        if (soapMessage.getMessageStyle() == SOAPMessageStyle.DOCUMENT_STYLE) {

            String soapBodyPart = ((SOAPDocumentStyleMessage)soapMessage).getBodyPart();

            if (soapBodyPart != null) {
                SOAPBodyElement soapBodyElement = null;
                if (soapVersion == SOAPVersion.SOAP11) {
                    soapBodyElement = wsdlDocument.createSOAP11BodyElement();
                } else {
                    soapBodyElement = wsdlDocument.createSOAP12BodyElement();
                }                
                List<String> partsNames = new ArrayList<String>();
                partsNames.add(soapBodyPart);
                soapBodyElement.setParts(partsNames);
                soapBodyElement.setUse("literal");
                bindingInputElement.setSOAPBodyElement(soapBodyElement);
            } else {
                SOAPBodyElement soapBodyElement = null;
                if (soapVersion == SOAPVersion.SOAP11) {
                    soapBodyElement = wsdlDocument.createSOAP11BodyElement();
                } else {
                    soapBodyElement = wsdlDocument.createSOAP12BodyElement();
                }   
                soapBodyElement.setUse("literal");
                bindingInputElement.setSOAPBodyElement(soapBodyElement);
            }

        } else {

            List<String> partNames = ((SOAPRpcStyleMessage)soapMessage).getBodyParts();
            String soapBodyNamespace = ((SOAPRpcStyleMessage)soapMessage).getSoapBodyNamespace();

            if (partNames.size() > 0) {
                SOAPBodyElement soapBodyElement = null;
                if (soapVersion == SOAPVersion.SOAP11) {
                    soapBodyElement = wsdlDocument.createSOAP11BodyElement();
                } else {
                    soapBodyElement = wsdlDocument.createSOAP12BodyElement();
                }   
                soapBodyElement.setParts(partNames);
                soapBodyElement.setUse("literal");
                soapBodyElement.setNamespace(soapBodyNamespace);
                bindingInputElement.setSOAPBodyElement(soapBodyElement);
            } else {
                SOAPBodyElement soapBodyElement = null;
                if (soapVersion == SOAPVersion.SOAP11) {
                    soapBodyElement = wsdlDocument.createSOAP11BodyElement();
                } else {
                    soapBodyElement = wsdlDocument.createSOAP12BodyElement();
                }   
                soapBodyElement.setUse("literal");
                soapBodyElement.setNamespace(soapBodyNamespace);
                bindingInputElement.setSOAPBodyElement(soapBodyElement);
            }

        }

        // now process the header parts. binding style is irrelevant. if header
        // is not mapped to a partname, i.e. message def wasn't imported via
        // wsdl 11, then we cannot emit header

        Collection<SOAPHeaderBlock> soapHeaderBlocks = soapMessage.getHeaders();

        QName inputMessageName = interfaceOperation.getInputMessageReference().getReferencedComponentName();
        for (SOAPHeaderBlock soapHeaderBlock : soapHeaderBlocks) {
            if (soapHeaderBlock.getPartName() == null) {
                continue;
            }
            SOAPHeaderElement soapHeaderElement = null;
            if (soapVersion == SOAPVersion.SOAP11) {
                soapHeaderElement = wsdlDocument.createSOAP11HeaderElement();
            } else {
                soapHeaderElement = wsdlDocument.createSOAP12HeaderElement();
            }            
            soapHeaderElement.setMessage(inputMessageName);
            soapHeaderElement.setPart(soapHeaderBlock.getPartName());
            soapHeaderElement.setUse("literal");
            bindingInputElement.addSOAPHeaderElement(soapHeaderElement);
        }

        bindingOperationElement.setInputElement(bindingInputElement);

        // create the optional output element, attach soap:body and optional headers

        if (bindingOperation.getOutputMessageReference() != null) {

            BindingOperationOutputElement bindingOutputElement = wsdlDocument
                    .createBindingOperationOutputElement();
            BindingMessageReference outputMessageReference = bindingOperation.getOutputMessageReference();

            if (outputMessageReference.getElementPolicy() != null) {
                PolicyExpression policy;
                try {
                    policy = outputMessageReference.getElementPolicy().getMergedExpression();
                } catch (PolicyException pe) {
                    throw new WSDLException("Error writing definition. " + pe);
                }
                writePolicy(policy, bindingOutputElement);
            }

            // process the body first. if this is a doc-lit style binding,
            // body contains zero or one part. if it's rpc-lit, it's a comma
            // separated list

            soapMessage = ((SOAPBindingMessageReference)outputMessageReference).getSoapMessage();

            if (soapMessage.getMessageStyle() == SOAPMessageStyle.DOCUMENT_STYLE) {

                String soapBodyPart = ((SOAPDocumentStyleMessage)soapMessage).getBodyPart();

                if (soapBodyPart != null) {
                    SOAPBodyElement soapBodyElement = null;
                    if (soapVersion == SOAPVersion.SOAP11) {
                        soapBodyElement = wsdlDocument.createSOAP11BodyElement();
                    } else {
                        soapBodyElement = wsdlDocument.createSOAP12BodyElement();
                    }   
                    List<String> partsNames = new ArrayList<String>();
                    partsNames.add(soapBodyPart);
                    soapBodyElement.setParts(partsNames);
                    soapBodyElement.setUse("literal");
                    bindingOutputElement.setSOAPBodyElement(soapBodyElement);
                } else {
                    SOAPBodyElement soapBodyElement = null;
                    if (soapVersion == SOAPVersion.SOAP11) {
                        soapBodyElement = wsdlDocument.createSOAP11BodyElement();
                    } else {
                        soapBodyElement = wsdlDocument.createSOAP12BodyElement();
                    }   
                    soapBodyElement.setUse("literal");
                    bindingOutputElement.setSOAPBodyElement(soapBodyElement);
                }

            } else {

                List<String> partNames = ((SOAPRpcStyleMessage)soapMessage).getBodyParts();
                String soapBodyNamespace = ((SOAPRpcStyleMessage)soapMessage).getSoapBodyNamespace();

                if (partNames.size() > 0) {
                    SOAPBodyElement soapBodyElement = null;
                    if (soapVersion == SOAPVersion.SOAP11) {
                        soapBodyElement = wsdlDocument.createSOAP11BodyElement();
                    } else {
                        soapBodyElement = wsdlDocument.createSOAP12BodyElement();
                    }   
                    soapBodyElement.setParts(partNames);
                    soapBodyElement.setUse("literal");
                    soapBodyElement.setNamespace(soapBodyNamespace);
                    bindingOutputElement.setSOAPBodyElement(soapBodyElement);
                } else {
                    SOAPBodyElement soapBodyElement = null;
                    if (soapVersion == SOAPVersion.SOAP11) {
                        soapBodyElement = wsdlDocument.createSOAP11BodyElement();
                    } else {
                        soapBodyElement = wsdlDocument.createSOAP12BodyElement();
                    }   
                    soapBodyElement.setUse("literal");
                    soapBodyElement.setNamespace(soapBodyNamespace);
                    bindingOutputElement.setSOAPBodyElement(soapBodyElement);
                }

            }

            // now process the header parts. binding style is irrelevant. if header
            // is not mapped to a partname, i.e. message def wasn't imported via
            // wsdl 11, then we cannot emit header

            soapHeaderBlocks = soapMessage.getHeaders();

            QName outputMessageName = interfaceOperation.getOutputMessageReference().getReferencedComponentName();
            for (SOAPHeaderBlock soapHeaderBlock : soapHeaderBlocks) {
                if (soapHeaderBlock.getPartName() == null) {
                    continue;
                }
                SOAPHeaderElement soapHeaderElement = null;
                if (soapVersion == SOAPVersion.SOAP11) {
                    soapHeaderElement = wsdlDocument.createSOAP11HeaderElement();
                } else {
                    soapHeaderElement = wsdlDocument.createSOAP12HeaderElement();
                }      
                soapHeaderElement.setMessage(outputMessageName);
                soapHeaderElement.setPart(soapHeaderBlock.getPartName());
                soapHeaderElement.setUse("literal");
                bindingOutputElement.addSOAPHeaderElement(soapHeaderElement);
            }

            bindingOperationElement.setOutputElement(bindingOutputElement);
        }

        // create a binding fault element for each fault defined by operation

        for (BindingFaultReference bindingFaultReference : bindingOperation.getFaultReferences()) {
            String faultName = bindingFaultReference.getFault().getFaultName().getLocalPart();
            SOAPFaultElement soapFaultElement = null;
            if (soapVersion == SOAPVersion.SOAP11) {
                soapFaultElement = wsdlDocument.createSOAP11FaultElement();
            } else {
                soapFaultElement = wsdlDocument.createSOAP12FaultElement();
            }
            soapFaultElement.setName(faultName);
            soapFaultElement.setUse("literal");
            BindingOperationFaultElement faultElement = wsdlDocument.createBindingOperationFaultElement();
            faultElement.setName(faultName);
            faultElement.setSOAPFaultElement(soapFaultElement);
            if (bindingFaultReference.getElementPolicy() != null) {
                PolicyExpression policy;
                try {
                    policy = bindingFaultReference.getElementPolicy().getMergedExpression();
                } catch (PolicyException pe) {
                    throw new WSDLException("Error writing definition. " + pe);
                }
                writePolicy(policy, faultElement);
            }

            bindingOperationElement.addFaultElement(faultElement);
        }
        
    }
    
    public static void writePolicy(PolicyExpression policyExpression, Element parent) throws WSDLException {
        parent.appendChild(parent.getOwnerDocument().adoptNode(policyExpression));
    }

}
